/*
 * Copyright (c)2005-2013, Haxe Foundation
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * Neither the name of the Haxe Foundation nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package haxe.xml;

class Printer {

	static public function print(xml:Xml) {
		var printer = new Printer();
		printer.writeNode(xml, "");
		return printer.output.toString();
	}

	var output:StringBuf;

	function new() {
		output = new StringBuf();
	}

	function writeNode(value:Xml, tabs:String) {
		switch (value.nodeType) {
			case Xml.CData:
				write(tabs + "<![CDATA[");
				write(StringTools.trim(value.nodeValue));
				write("]]>");
				newline();
			case Xml.Comment:
				var commentContent:String = value.nodeValue;
				commentContent = ~/[\n\r\t]+/g.replace(commentContent, "");
				write(tabs);
				write(StringTools.trim(commentContent));
				newline();
			case Xml.Document:
				for (child in value) {
					writeNode(child, tabs);
				}
			case Xml.Element:
				write(tabs + "<");
				write(value.nodeName);
				for (attribute in value.attributes()) {
					write(" " + attribute + "=\"");
					write(value.get(attribute));
					write("\"");
				}
				if (hasChildren(value)) {
					write(">");
					newline();
					for (child in value) {
						writeNode(child, tabs + "\t");
					}
					write(tabs + "</");
					write(value.nodeName);
					write(">");
					newline();
				} else {
					write(" />");
					newline();
				}
		case Xml.PCData:
			var nodeValue:String = StringTools.trim(value.nodeValue);
				if (nodeValue.length != 0) {
					write(tabs + nodeValue);
					newline();
				}
			}
	}

	inline function write(input:String) {
		output.add(input);
	}

	inline function newline() {
		output.add("\n");
	}

	function hasChildren(value:Xml):Bool {
		for (child in value) {
			switch (child.nodeType) {
				case Xml.Element:
					return true;
				case Xml.CData, Xml.Comment, Xml.PCData:
					if (StringTools.ltrim(child.nodeValue).length != 0) {
						return true;
					}
			}
		}
		return false;
	}
}